"""test sql_db

:copyright: Copyright (c) 2025 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

_PATH = "test.sqlite3"


def test_1():
    # Must be first
    from pykern.pkcollections import PKDict
    from pykern import pkunit, pkdebug

    with pkunit.save_chdir_work() as d:
        m = _meta(d)
        _validate_schema(m)
        _inserts(m)
        _selects(m)
        _deletes(m)


def _deletes(meta):
    from pykern import pkunit

    with meta.session() as s:
        pkunit.pkeq(len(s.select("t2", dict(a_bigint=333)).all()), 1)
        s.delete("t2", dict(a_bigint=333))
        pkunit.pkeq(len(s.select("t2", dict(a_bigint=333)).all()), 0)


def _inserts(meta):
    with meta.session() as s:
        r = s.insert("t1", a_name="William", a_int=1, a_text="war and peace")
        s.insert(
            "T2",
            t1_id=r.t1_id,
            a_float=3,
            a_double=1e1000,
            a_bigint=int(1e11),
        )
        r = s.insert(
            "T1", dict(a_name="Mildred", a_int=333, a_text="hitch hiker's guide")
        )
        s.insert("t2", dict(t1_id=r.t1_id, a_float=2, a_bigint=222))
        s.insert(s.t.t2, dict(t1_id=r.t1_id, a_float=3, a_bigint=333))


def _meta(dir_path):
    from pykern.pkcollections import PKDict
    from pykern import sql_db

    return sql_db.Meta(
        uri=_uri(dir_path),
        schema=PKDict(
            T1=PKDict(
                t1_id="primary_id 1",
                a_name="str 64",
                a_text="text",
                a_int="int 32",
                created="datetime index",
                unique=(("a_name", "a_text"),),
            ),
            t2=PKDict(
                t2_id="primary_id 2",
                t1_id="primary_id",
                a_float="float 32",
                a_double="float 64 nullable",
                a_bigint="int 64 unique",
                index=(("t1_id", "t2_id"),),
            ),
            t3=PKDict(
                t3_name="str 4 primary_key",
            ),
            t4=PKDict(
                t3_name="str 4 primary_key foreign",
                t4_pk2="str 4 primary_key",
            ),
            t5=PKDict(
                t3_name="str 4 primary_key",
                t4_pk2="str 4 primary_key",
                foreign=((("t3_name", "t4_pk2"), ("t4.t3_name", "t4.t4_pk2")),),
            ),
            t6=PKDict(
                t2_id="primary_id primary_key",
                t1_id="primary_id primary_key",
                t6_val="bool",
            ),
        ),
    )


def _selects(meta):
    from pykern import pkunit, pkdebug
    import sqlalchemy

    with meta.session() as s:
        r = s.select_one("t1", where=dict(a_name="Mildred"))
        pkunit.pkeq("hitch hiker's guide", r.a_text)
        t1, t2 = s.t.t1, s.t.t2
        r = s.execute(
            sqlalchemy.select(
                t1,
                t2,
            )
            .join(
                t1,
                t2.c.a_bigint == t1.c.a_int,
            )
            .where(
                t1.c.a_int == 333,
            ),
        )


def _uri(d):
    from pykern import sql_db

    return sql_db.sqlite_uri(d.join(_PATH))


def _validate_schema(meta):
    import re, subprocess
    from pykern import pkunit

    # Need to reorder, because sqlite does not keep a consistent order of index output
    t = []
    i = []
    for l in re.split(
        "\s*\n\s*",
        # Assumes in same directory
        subprocess.check_output(["sqlite3", _PATH, ".schema"], text=True),
    ):
        if len(l) <= 0:
            continue
        l += "\n"
        if re.search("^CREATE(?: UNIQUE)? INDEX", l):
            i.append(l)
        else:
            t.append(l)
    pkunit.file_eq("schema.txt", "".join(t + sorted(i)))
    pkunit.pkeq(["t1", "t2", "t3", "t4", "t5", "t6"], sorted(meta.t.keys()))
    pkunit.pkeq(meta.t.t1, meta.table("t1"))



"""test fconf

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_parser():
    from pykern import pkunit, fconf, pkio, pkjson
    from pykern.pkcollections import PKDict
    from pykern.pkdebug import pkdp

    for d in pkunit.case_dirs():
        p = fconf.Parser(
            pkio.sorted_glob("*.py") + pkio.sorted_glob("*.yml"),
            base_vars=(
                PKDict(fconf_test=PKDict(var1=1234)) if d.basename == "1" else None
            ),
        )
        r = p.result
        r.macros = [f'{p.name}({",".join(p.params)})' for p in p.macros.values()]
        pkjson.dump_pretty(r, filename="res.json")


def test_parse_all():
    from pykern import pkunit, fconf, pkjson

    d = pkunit.work_dir().join("parseall").ensure(dir=1)
    d.join("1.yml").write("---\npi: 3.14\ne: 2.7\n")
    d.join("2.yaml").write("---\npi: ${e}\n")
    d.join("3.yml").write("---\ne: 1.6\n")
    d.join("not-seen.yml").write("---\nnot-seen: false\n")
    a = fconf.parse_all(d, glob="?")
    pkjson.dump_pretty(a, filename=d.join("res.json"))
    pkunit.pkok("not-seen" not in a, "not-seen got parsed values={}", a)
    pkunit.pkeq(2.7, a.pi)
    pkunit.pkeq(1.6, a.e)



"""Where external resources are stored

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Root module: avoid importing modules which import pkconfig
from pykern import pkconst
from pykern import pkinspect
from pykern import pkio
import errno
import glob
import importlib
import os.path
import sys

from pykern.pkdebug import pkdp

if sys.version_info < (3, 10):
    # There's a bug resources.files() in 3.9.15 so need this.
    # Also works with older versions than 3.9
    import pkg_resources

    def _resource_filename(package, path):
        return pkg_resources.resource_filename(
            package,
            os.path.join(pkconst.PACKAGE_DATA, path),
        )

else:
    import importlib.resources

    def _resource_filename(package, path):
        return str(
            importlib.resources.files(package).joinpath(pkconst.PACKAGE_DATA, path)
        )


def file_path(relative_filename, caller_context=None, packages=None):
    """Return the path to the resource

    Args:
        relative_filename (str): file name relative to package_data directory.
        caller_context (object): Any object from which to get the `root_package`
        packages (List[str]): Packages to search.

    Returns:
        py.path: absolute path of the resource file
    """
    return pkio.py_path(filename(relative_filename, caller_context, packages))


def filename(relative_filename, caller_context=None, packages=None):
    """Return the filename to the resource

    Args:
        relative_filename (str): file name relative to package_data directory.
        caller_context (object): Any object from which to get the `root_package`
        packages (List[str]): Packages to search.

    Returns:
        str: absolute path of the resource file
    """
    assert not os.path.isabs(
        relative_filename
    ), "must not be an absolute file name={}".format(relative_filename)
    a = []
    for f, p in _files(relative_filename, caller_context, packages):
        a.append(p)
        if os.path.exists(f):
            return f
    _raise_no_file_found(a, relative_filename)


def glob_paths(relative_path, caller_context=None, packages=None):
    """Find all paths that match the relative path in all packages

    Args:
        relative_path(str): Path relative to package_data directory.
        caller_context (object): Any object from which to get the `root_package`.
        packages (List[str]): Packages to search.
    Returns:
        py.path: absolute paths of the matched files
    """
    r = []
    for f, p in _files(relative_path, caller_context, packages):
        r.extend(glob.glob(f))
    return [pkio.py_path(f) for f in r]


def _files(path, caller_context, packages):

    if caller_context and packages:
        raise ValueError(
            f"Use only one of caller_context={caller_context} and packages={packages}",
        )
    for p in list(
        map(
            lambda m: pkinspect.root_package(importlib.import_module(m)),
            packages
            or [
                pkinspect.root_package(
                    caller_context if caller_context else pkinspect.caller_module()
                )
            ],
        )
    ):
        yield (
            _resource_filename(p, path),
            p,
        )


def _raise_no_file_found(packages, path):
    msg = f"unable to locate in packages={packages}"
    if "__main__" in packages:
        msg += "; do not call module as a program"
    raise IOError(errno.ENOENT, msg, path)



def checking_prints_at_function_level():
    from pykern.pkdebug import pkdp

    pkdp("this pkdp should be found")
    print("this print should be found")
    "#" + print("should be found")
    int(pkdp(1))



"""manipulate Excel spreadsheets

:copyright: Copyright (c) 2025 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp
import pandas
import pykern.pkcli
import pykern.pkio


def to_csv(xlsx_path, sheet=None, csv_path=None):
    """Dump sheets from xlsx_path

    Args:
        xlsx_path (str or py.path): what to parse
        sheet (str or int): dump a specfic sheet [None: all]
        csv_path (str or py.path): where to write [base#N.csv]
    Returns:
        tuple: list of csv files created
    """
    return ToCSV(xlsx_path, sheet, csv_path).result


class ToCSV:

    def __init__(self, xlsx_path, sheet, csv_path):
        if csv_path is not None:
            csv_path = pykern.pkio.py_path(csv_path)
        self._xlsx_path = pykern.pkio.py_path(xlsx_path)
        self._xlsx = pandas.ExcelFile(str(self._xlsx_path))
        self._csv_path = csv_path or self._xlsx_path
        self.result = (
            tuple(self._multiple()) if sheet is None else (self._one(sheet, csv_path),)
        )

    def _multiple(self):
        for i in range(len(self._xlsx.sheet_names)):
            yield self._one(i, None)

    def _one(self, sheet, csv_path):
        if csv_path is None:
            csv_path = self._csv_path.new(
                purebasename=f"{self._csv_path.purebasename}#{sheet}",
                ext=".csv",
            )
        try:
            sheet = int(sheet)
        except ValueError:
            pass
        d = self._xlsx.parse(index_col=None, sheet_name=sheet)
        d.columns = d.columns.map(lambda c: "" if "Unnamed" in str(c) else str(c))
        d.to_csv(
            str(csv_path),
            encoding="utf-8",
            index=False,
            lineterminator="\r\n",
        )
        return csv_path



"""test pkcli.xlsx

:copyright: Copyright (c) 2025 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_to_csv():
    from pykern import pkunit

    for d in pkunit.case_dirs():
        from pykern.pkcli import xlsx

        xlsx.to_csv("in.xlsx")



"""test ci

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_all():
    from pykern import pkunit
    from pykern.pkcli import ci

    for d in pkunit.case_dirs():
        with pkunit.ExceptToFile():
            getattr(ci, d.basename.split("-")[0])()



""":mod:`private` package

:copyright: Copyright (c) 2024 James Bond.  All Rights Reserved.
:license: PROPRIETARY AND CONFIDENTIAL. See LICENSE file for details.
"""
import importlib.metadata

try:
    # We only have a version once the package is installed.
    __version__ = importlib.metadata.version("private")
except importlib.metadata.PackageNotFoundError:
    pass
